from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from dissect.target.plugins.os.default._os import DefaultOSPlugin
from dissect.target.plugins.os.unix.linux._os import LinuxPlugin
from dissect.target.plugins.os.unix.linux.fortios._os import FortiOSPlugin
from dissect.target.target import Target

from acquire.acquire import PROFILES, _get_modules_for_profile

if TYPE_CHECKING:
    from dissect.target.plugin import OSPlugin


@pytest.mark.parametrize(
    argnames=("os_plugin", "expected_value"),
    argvalues=[
        (
            FortiOSPlugin,
            ["Etc", "Boot", "Home", "SSH", "Var"],
        ),
        (
            LinuxPlugin,
            ["Etc", "Boot", "Home", "SSH", "Var"],
        ),
        (
            DefaultOSPlugin,
            [],
        ),
    ],
)
def test_profile_selection_linux(os_plugin: OSPlugin, expected_value: list[str]) -> None:
    target = Target()
    target._os_plugin = os_plugin
    target.apply()

    assert (
        list(_get_modules_for_profile(target, "minimal", PROFILES, "No collection set for OS '%s' with profile '%s'"))
        == expected_value
    )



from __future__ import annotations

import csv
import ctypes
import gzip
import io
import threading
from logging import Filter, getLogger
from queue import Empty, Queue
from typing import TYPE_CHECKING

from acquire.dynamic.windows.exceptions import OpenProcessError
from acquire.dynamic.windows.ntdll import (
    NtQueryInformationFile,
    NtQueryObject,
    NtQuerySystemInformation,
    NtStatusCode,
    close_handle,
)
from acquire.dynamic.windows.types import (
    BOOL,
    DWORD,
    FILE_INFORMATION_CLASS,
    HANDLE,
    IO_STATUS_BLOCK,
    OBJECT_INFORMATION_CLASS,
    PHANDLE,
    PUBLIC_OBJECT_TYPE_INFORMATION,
    SYSTEM_HANDLE_INFORMATION_EX,
    SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX,
    SYSTEM_INFORMATION_CLASS,
    ULONG,
    DuplicateHandleFlags,
    ErrorCode,
    FileNameInformationFactory,
    Handle,
    ProcessAccess,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from logging import LogRecord

log = getLogger(__name__)

OpenProcessToken = ctypes.windll.advapi32.OpenProcessToken
OpenProcessToken.argtypes = [HANDLE, DWORD, PHANDLE]

kernel32 = ctypes.windll.kernel32
OpenProcess = kernel32.OpenProcess
OpenProcess.restype = HANDLE

DuplicateHandle = kernel32.DuplicateHandle
DuplicateHandle.argtypes = [HANDLE, HANDLE, HANDLE, ctypes.POINTER(HANDLE), DWORD, BOOL, DWORD]

GetLastError = kernel32.GetLastError
SetLastError = kernel32.SetLastError
GetCurrentProcessId = kernel32.GetCurrentProcessId


class DuplicateFilter(Filter):
    def __init__(self) -> None:
        super().__init__()
        self.msgs = set()

    def filter(self, record: LogRecord) -> bool:
        msg = record.getMessage()
        if show := msg not in self.msgs:
            self.msgs.add(msg)

        return show


def get_handle_type_info(handle: SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX) -> str | None:
    """Return type of handle.

    Args:
        handle: handle for which to return the type information.

    Raises:
        RuntimeError: Raised when the result of the object query is unknown (other than SUCCESS, LENGTH MISMATCH or
        INVALID).
    """
    public_object_type_information = PUBLIC_OBJECT_TYPE_INFORMATION()
    size = DWORD(ctypes.sizeof(public_object_type_information))
    while True:
        result = NtQueryObject(
            handle,
            OBJECT_INFORMATION_CLASS.ObjectTypeInformation,
            ctypes.byref(public_object_type_information),
            size,
            None,
        )

        if result == NtStatusCode.STATUS_SUCCESS:
            return public_object_type_information.name
        if result == NtStatusCode.STATUS_INFO_LENGTH_MISMATCH:
            size = DWORD(size.value * 4)
            ctypes.resize(public_object_type_information, size.value)
        elif result == NtStatusCode.STATUS_INVALID_HANDLE:
            return None
        else:
            raise RuntimeError(hex(result))


def open_process(pid: int) -> int:
    """Obtain a handle for the given PID.

    More info: https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-openprocess

    Args:
        pid: integer that represents the process ID.

    Raises:
        OpenProcessError: Raies when the System Idle Process, the System Process or one of the CSRSS processes are tried
        to be opened.
    """
    SetLastError(0)

    h_process = OpenProcess(
        ProcessAccess.PROCESS_DUP_HANDLE,
        False,
        pid,
    )

    error = GetLastError()
    if error in [ErrorCode.ERROR_INVALID_PARAMETER, ErrorCode.ERROR_ACCESS_DENIED]:
        raise OpenProcessError(
            f"Likely tried opening the System Idle Process, the System Process or one of the Client Server Run-Time"
            f"Subsystem (CSRSS) processes [pid: {pid}]"
        )

    # No valid handle could be obtained, display the error code
    if h_process == 0:
        raise OpenProcessError(f"OpenProcess Error: 0x{error:x} [pid: {pid}]")

    return h_process


def _get_file_name_thread(h_file: HANDLE, queue: Queue) -> None:
    iob = IO_STATUS_BLOCK()
    file_name_information = FileNameInformationFactory()
    file_name = None

    while True:
        result = NtQueryInformationFile(
            h_file,
            ctypes.byref(iob),
            ctypes.byref(file_name_information),
            ULONG(ctypes.sizeof(file_name_information)),
            FILE_INFORMATION_CLASS.FileNameInformation,
        )

        if result == NtStatusCode.STATUS_BUFFER_OVERFLOW:
            file_name_information = FileNameInformationFactory(file_name_information.FileNameLength)
        elif result == NtStatusCode.STATUS_SUCCESS:
            file_name = file_name_information.FileName
            break
        else:
            # Multiple StatusCodes can be observed. In almost all cases FileNameLength is 0. Breaking for now
            break

    queue.put(file_name)


def get_handle_name(pid: int, handle: SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX) -> str | None:
    """Return handle name."""
    remote = pid != GetCurrentProcessId()

    if remote:
        try:
            h_remote = open_process(pid)
        except OpenProcessError as e:
            log.error(e)  # noqa: TRY400
            return None
        try:
            handle = duplicate_handle(h_remote, handle)
        except RuntimeError:
            close_handle(h_remote)
            return None

    # Use threading to try (max a second) to get the handle name, since it might hang
    queue = Queue()
    thread = threading.Thread(target=_get_file_name_thread, args=(handle, queue))
    thread.daemon = True
    thread.start()
    thread.join(1.0)

    result = None
    if not thread.is_alive():
        try:
            result = queue.get_nowait()
        except Empty:
            pass

    return result


def get_handles() -> Iterator[Handle]:
    """Returns all handles of a target."""
    system_handle_information = SYSTEM_HANDLE_INFORMATION_EX()
    size = DWORD(ctypes.sizeof(system_handle_information))
    duplicate_filter = DuplicateFilter()
    log.addFilter(duplicate_filter)

    while True:
        result = NtQuerySystemInformation(
            SYSTEM_INFORMATION_CLASS.SystemExtendedHandleInformation,
            ctypes.byref(system_handle_information),
            size,
            ctypes.byref(size),
        )

        if result == NtStatusCode.STATUS_SUCCESS:
            break
        elif result == NtStatusCode.STATUS_INFO_LENGTH_MISMATCH:
            size = DWORD(size.value * 4)
            ctypes.resize(system_handle_information, size.value)
        else:
            raise RuntimeError(hex(result))

    p_handles = ctypes.cast(
        system_handle_information.Handles,
        ctypes.POINTER(SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX * system_handle_information.NumberOfHandles),
    )
    for handle in p_handles.contents:
        try:
            handle_type = get_handle_type_info(handle.HandleValue)
            handle_name = get_handle_name(handle.UniqueProcessId, handle.HandleValue)

            if not handle_name:
                continue

            yield Handle(handle, handle_type, handle_name)
        except Exception as handle_error:
            log.error("An error occurred while parsing handle, skipping handle. Error: %s", handle_error)  # noqa: TRY400

    log.removeFilter(duplicate_filter)


def duplicate_handle(h_process: int, handle: SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX) -> HANDLE:
    """Create duplicate handle.

    When the source handle is in use by another program, one needs to create a duplicate handle in order to have full
    control of that handle. This prevents performing operations on the source handle that might have been closed by
    the other program.
    """
    h_dup = HANDLE()
    SetLastError(0)
    result = DuplicateHandle(
        h_process,
        handle,
        kernel32.GetCurrentProcess(),
        ctypes.byref(h_dup),
        0,
        False,
        DuplicateHandleFlags.DUPLICATE_SAME_ACCESS,
    )
    if result == 0:
        raise RuntimeError

    return h_dup


def serialize_handles_into_csv(rows: Iterator[Handle], compress: bool = True) -> bytes:
    """Serialize handle data into a csv.

    Serialize provided rows into normal or gzip-compressed CSV, and return a tuple
    containing the result bytes.
    """
    raw_buffer = io.BytesIO()

    buffer = gzip.GzipFile(fileobj=raw_buffer, mode="wb") if compress else raw_buffer

    with io.TextIOWrapper(buffer, encoding="utf-8") as wrapper:
        csv_writer = None
        for i, row in enumerate(rows):
            if i == 0:
                csv_writer = csv.DictWriter(wrapper, fieldnames=row.dictionary.keys())
                csv_writer.writeheader()

            csv_writer.writerow(row.dictionary)

    return raw_buffer.getvalue()



from __future__ import annotations

from acquire.outputs.dir import DirectoryOutput
from acquire.outputs.tar import TAR_COMPRESSION_METHODS, TarOutput
from acquire.outputs.zip import ZIP_COMPRESSION_METHODS, ZipOutput

__all__ = ["DirectoryOutput", "TarOutput", "ZipOutput"]

OUTPUTS = {"tar": TarOutput, "dir": DirectoryOutput, "zip": ZipOutput}

COMPRESSION_METHODS = {*TAR_COMPRESSION_METHODS, *ZIP_COMPRESSION_METHODS}



from __future__ import annotations

import io
import shutil
import stat
import zipfile
from datetime import datetime, timezone
from typing import TYPE_CHECKING, BinaryIO

from acquire.crypt import EncryptedStream
from acquire.outputs.base import Output

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.filesystem import FilesystemEntry

ZIP_COMPRESSION_METHODS = {"deflate": zipfile.ZIP_DEFLATED, "bzip2": zipfile.ZIP_BZIP2, "lzma": zipfile.ZIP_LZMA}


class ZipOutput(Output):
    """Zip archive acquire output format. Output can be compressed and/or encrypted.

    Args:
        path: The path to write the zip archive to.
        compress: Whether to compress the zip archive.
        compression_method: Compression method to use (Default: Deflate). Supports "deflate", "bzip2", "lzma".
        encrypt: Whether to encrypt the zip archive.
        public_key: The RSA public key to encrypt the header with.
    """

    def __init__(
        self,
        path: Path,
        compress: bool = False,
        compression_method: str = "deflate",
        encrypt: bool = False,
        public_key: bytes | None = None,
    ) -> None:
        ext = ".zip" if ".zip" not in path.suffixes else ""

        if encrypt:
            ext += ".enc"

        self._fh = None
        self.path = path.with_suffix(path.suffix + ext)

        if compress:
            self.compression = ZIP_COMPRESSION_METHODS.get(compression_method, zipfile.ZIP_DEFLATED)
        else:
            self.compression = zipfile.ZIP_STORED

        if encrypt:
            self._fh = EncryptedStream(self.path.open("wb"), public_key)
            self.archive = zipfile.ZipFile(self._fh, mode="w", compression=self.compression, allowZip64=True)
        else:
            self.archive = zipfile.ZipFile(self.path, mode="w", compression=self.compression, allowZip64=True)

    def write(
        self,
        output_path: str,
        fh: BinaryIO,
        entry: FilesystemEntry | Path | None = None,
        size: int | None = None,
    ) -> None:
        """Write a filesystem entry or file-like object to a zip file.

        Args:
            output_path: The path of the entry in the output format.
            fh: The file-like object of the entry to write.
            entry: The optional filesystem entry of the entry to write.
            size: The optional file size in bytes of the entry to write.
        """
        lstat = None
        size = size or getattr(fh, "size", None)

        if size is None and fh.seekable():
            offset = fh.tell()
            fh.seek(0, io.SEEK_END)
            size = fh.tell()
            fh.seek(offset)

        info = zipfile.ZipInfo()
        info.filename = output_path
        info.file_size = size or 0
        info.compress_type = self.compression

        if entry:
            info.external_attr = self._get_external_attr(entry)

            if entry.is_symlink():
                # System which created ZIP archive, 3 = Unix; 0 = Windows
                # Windows does not have symlinks, so this must be a unixoid system
                info.create_system = 3

            lstat = entry.lstat()
            if lstat:
                # Python zipfile module does not support timestamps before 1980
                dt = datetime.fromtimestamp(lstat.st_mtime, tz=timezone.utc)
                year = max(dt.year, 1980)
                info.date_time = (year, dt.month, dt.day, dt.hour, dt.minute, dt.second)

        with self.archive.open(info, "w") as zfh:
            shutil.copyfileobj(fh, zfh)

    def close(self) -> None:
        """Closes the archive file."""
        self.archive.close()
        if self._fh:
            self._fh.close()

    def _get_external_attr(self, entry: FilesystemEntry) -> int:
        """Return the appropriate external attributes of the entry."""
        # The Python zipfile module accepts the 16-bit "Mode" field (that stores st_mode field from
        # struct stat, containing user/group/other permissions, setuid/setgid and symlink info, etc) of the
        # ASi extra block for Unix as bits 16-31 of the external_attr
        unix_st_mode = stat.S_IFREG

        if entry.is_symlink():
            unix_st_mode = stat.S_IFLNK
        elif entry.is_dir():
            unix_st_mode = stat.S_IFDIR

        return (
            unix_st_mode
            | stat.S_IRUSR
            | stat.S_IWUSR
            | stat.S_IXUSR
            | stat.S_IRGRP
            | stat.S_IWGRP
            | stat.S_IXGRP
            | stat.S_IROTH
            | stat.S_IWOTH
            | stat.S_IXOTH
        ) << 16



from __future__ import annotations

import time
from threading import Thread
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from argparse import Namespace


class GUIError(RuntimeError):
    pass


class GUI:
    _instance = None
    _progress = 0
    _closed = False
    _shard = 0

    thread = None
    folder = None
    files = None
    ready = False
    auto_upload = None
    upload_available = False

    def __new__(cls, flavour: str | None = None, upload_available: bool = False):
        # singleton+factory pattern
        if cls._instance is None:
            cls = Stub
            if str(flavour).lower() == "windows":
                # create a basic Win32 GUI
                from acquire.gui.win32 import Win32  # noqa: PLC0415

                cls = Win32
            GUI._instance = super().__new__(cls)
            GUI._instance.upload_available = upload_available
        return GUI._instance

    @classmethod
    def gui(cls) -> GUI:
        """Returns the instance of the GUI."""
        return cls._instance

    @property
    def shard(self) -> int:
        """Returns the shard of the progress bar."""
        return self._shard

    @shard.setter
    def shard(self, shard: int) -> None:
        """Sets the shard of the progress bar."""
        # Use this to 'refine' progress bar (i.e. assign a shard)
        if shard > 100 or shard < 0:
            raise GUIError("Shards have to be between 0-100")
        self._shard = shard

    def wait_for_start(self, args: Namespace) -> tuple[str, list[str], bool, bool]:
        """Starts GUI thread and waits for start button to be clicked."""

        def gui_thread() -> None:
            self.show()

        GUI.thread = Thread(target=gui_thread)
        GUI.thread.start()
        while not self.ready and not self._closed:
            time.sleep(1)
        return self.folder, self.files, self.auto_upload, self._closed

    def message(self, message: str) -> None:
        """Starts GUI thread and waits for start button to be clicked."""
        raise NotImplementedError

    def wait_for_quit(self) -> None:
        """Closes the GUI and waits for the thread to join."""
        GUI._instance.quit()
        GUI.thread.join()
        self._closed = True

    def show(self) -> None:
        """Subclass needs to implement this."""
        raise NotImplementedError

    def finish(self) -> None:
        """Finishes the progress bar and closes the GUI."""
        raise NotImplementedError


class Stub(GUI):
    """Minimal GUI implementation."""

    def message(self, message: str) -> None:
        pass

    def wait_for_start(self, args: Namespace) -> tuple[str, list[str], bool, bool]:
        return args.output, [], args.auto_upload, False

    def wait_for_quit(self) -> None:
        pass

    def finish(self) -> None:
        pass



from __future__ import annotations

import time
from ctypes import (
    HRESULT,
    POINTER,
    WINFUNCTYPE,
    Structure,
    WinDLL,
    WinError,
    byref,
    c_int,
    c_int64,
    c_uint,
    c_void_p,
    cast,
    create_string_buffer,
    create_unicode_buffer,
    get_last_error,
    sizeof,
    string_at,
)
from ctypes import wintypes as w
from pathlib import Path
from typing import Any

from acquire.gui import GUI, GUIError


def _winerror(result: int, *args) -> Any:
    if not result:
        raise WinError(get_last_error())
    return result


def LOWORD(dword: w.DWORD) -> w.DWORD:
    return dword & 0x0000FFFF


def HIWORD(dword: w.DWORD) -> w.DWORD:
    return dword >> 16


LRESULT = c_int64
HCURSOR = c_void_p
WNDPROC = WINFUNCTYPE(LRESULT, w.HWND, w.UINT, w.WPARAM, w.LPARAM)
BFFCALLBACK = WINFUNCTYPE(c_int, w.HWND, w.UINT, w.LPARAM, w.LPARAM)
CW_USEDEFAULT = -0x80000000
IDI_APPLICATION = w.LPCWSTR(0x7F00)
CS_VREDRAW = 0x0001
CS_HREDRAW = 0x0002
IDC_ARROW = w.LPCWSTR(0x7F00)
WHITE_BRUSH = 0
SW_SHOWNORMAL = 1
WM_DESTROY = 0x0002
WM_ENABLE = 0x000A
WM_PAINT = 0x000F
WM_CLOSE = 0x0010
WM_SETFONT = 0x0030
WM_COMMAND = 0x0111
WM_USER = 0x0400
DT_SINGLELINE = 32
DT_CENTER = 1
DT_VCENTER = 4
WS_CHILD = 0x40000000
WS_VISIBLE = 0x10000000
WS_BORDER = 0x00800000
WS_OVERLAPPEDWINDOW = 0xCF0000
WS_DISABLED = 0x08000000
BS_PUSHBUTTON = 0
BS_CHECKBOX = 2
BS_AUTOCHECKBOX = 3
BS_CENTER = 300
BS_FLAT = 0x8000
ES_PASSWORD = 32
ES_WANTRETURN = 4096
EM_SETPASSWORDCHAR = 204
PBM_SETRANGE = WM_USER + 1
PBM_SETPOS = WM_USER + 2
PBM_DELTAPOS = WM_USER + 3
PBM_SETSTEP = WM_USER + 4
PBM_STEPIT = WM_USER + 5
PBM_SETRANGE32 = WM_USER + 6
PBS_SMOOTH = 0x01
BN_CLICKED = 0
BM_SETCHECK = 241
WM_CTLCOLORSTATIC = 312
SS_LEFT = 0
FNERR_BUFFERTOOSMALL = 0x3003

OFN_ALLOWMULTISELECT = 0x00000200
OFN_PATHMUSTEXIST = 0x00000800
OFN_FILEMUSTEXIST = 0x00001000
OFN_EXPLORER = 0x00080000


class WNDCLASSW(Structure):
    _fields_ = (
        ("style", w.UINT),
        ("lpfnWndProc", WNDPROC),
        ("cbClsExtra", c_int),
        ("cbWndExtra", c_int),
        ("hInstance", w.HINSTANCE),
        ("hIcon", w.HICON),
        ("hCursor", HCURSOR),
        ("hbrBackground", w.HBRUSH),
        ("lpszMenuName", w.LPCWSTR),
        ("lpszClassName", w.LPCWSTR),
    )


class PAINTSTRUCT(Structure):
    _fields_ = (
        ("hdc", w.HDC),
        ("fErase", w.BOOL),
        ("rcPaint", w.RECT),
        ("fRestore", w.BOOL),
        ("fIncUpdate", w.BOOL),
        ("rgbReserved", w.BYTE * 32),
    )


class INITCOMMONCONTROLSEX(Structure):
    _fields_ = (
        ("dwSize", w.DWORD),
        ("dwICC", w.DWORD),
    )


class BROWSEINFOA(Structure):
    _fields_ = (
        ("hwndOwner", w.HWND),
        ("pidlRoot", w.LPVOID),
        ("pszDisplayName", w.LPSTR),
        ("lpszTitle", w.LPCSTR),
        ("ulFlags", c_uint),
        ("lpfn", BFFCALLBACK),
        ("lParam", w.LPARAM),
        ("iImage", c_int),
    )


class SHITEMID(Structure):
    _fields_ = (
        ("cb", w.USHORT),
        ("abID", w.BYTE),
    )


class ITEMIDLIST(Structure):
    _fields_ = (("mkid", SHITEMID),)


class OPENFILENAMEW(Structure):
    _fields_ = (
        ("lStructSize", w.DWORD),
        ("hwndOwner", w.HWND),
        ("hInstance", w.HINSTANCE),
        ("lpstrFilter", w.LPCWSTR),
        ("lpstrCustomFilter", w.LPWSTR),
        ("nMaxCustFilter", w.DWORD),
        ("nFilterIndex", w.DWORD),
        ("lpstrFile", w.LPWSTR),
        ("nMaxFile", w.DWORD),
        ("lpstrFileTitle", w.LPWSTR),
        ("nMaxFileTitle", w.DWORD),
        ("lpstrInitialDir", w.LPCWSTR),
        ("lpstrTitle", w.LPCWSTR),
        ("Flags", w.DWORD),
        ("nFileOffset", w.WORD),
        ("nFileExtension", w.WORD),
        ("lpstrDefExt", w.LPWSTR),
        ("lCustData", w.LPARAM),
        ("lpfnHook", w.LPVOID),
        ("lpTemplateName", w.LPCSTR),
        ("pvReserved", w.LPVOID),
        ("dwReserved", w.DWORD),
        ("FlagsEx", w.DWORD),
    )


comdlg32 = WinDLL("comdlg32", use_last_error=True)
comdlg32.GetOpenFileNameW.argtypes = (POINTER(OPENFILENAMEW),)
comdlg32.GetOpenFileNameW.restype = w.BOOL
comdlg32.CommDlgExtendedError.argtypes = ()
comdlg32.CommDlgExtendedError.restype = w.DWORD

kernel32 = WinDLL("kernel32", use_last_error=True)
kernel32.GetModuleHandleW.argtypes = (w.LPCWSTR,)
kernel32.GetModuleHandleW.restype = w.HMODULE
kernel32.GetModuleHandleW._winerror = _winerror

user32 = WinDLL("user32", use_last_error=True)
user32.CreateWindowExW.argtypes = (
    w.DWORD,
    w.LPCWSTR,
    w.LPCWSTR,
    w.DWORD,
    c_int,
    c_int,
    c_int,
    c_int,
    w.HWND,
    w.HMENU,
    w.HINSTANCE,
    w.LPVOID,
)
user32.CreateWindowExW.restype = w.HWND
user32.CreateWindowExW._winerror = _winerror
user32.DefWindowProcW.argtypes = (
    w.HWND,
    w.UINT,
    w.WPARAM,
    w.LPARAM,
)
user32.DefWindowProcW.restype = LRESULT
user32.DestroyWindow.argtypes = (w.HWND,)
user32.DestroyWindow.restype = w.BOOL
user32.DestroyWindow._winerror = _winerror
user32.EnableWindow.argtypes = (
    w.HWND,
    w.BOOL,
)
user32.EnableWindow.restype = w.BOOL
user32.EnableWindow._winerror = _winerror
user32.MessageBoxA.argtypes = (
    w.HWND,
    w.LPCSTR,
    w.LPCSTR,
    c_uint,
)
user32.MessageBoxA.restype = c_int
user32.MessageBoxA._winerror = _winerror
user32.SetWindowTextA.argtypes = (
    w.HWND,
    w.LPCSTR,
)
user32.SetWindowTextA.restype = w.BOOL
user32.SetWindowTextA._winerror = _winerror
SendMessage = user32.SendMessageA
SendMessage.argtypes = (w.HWND, w.UINT, w.WPARAM, w.LPARAM)
SendMessage.restype = c_void_p

gdi32 = WinDLL("gdi32", use_last_error=True)
gdi32.CreateFontA.argtypes = (
    c_int,
    c_int,
    c_int,
    c_int,
    c_int,
    w.DWORD,
    w.DWORD,
    w.DWORD,
    w.DWORD,
    w.DWORD,
    w.DWORD,
    w.DWORD,
    w.DWORD,
    w.LPCSTR,
)
gdi32.CreateFontA.restype = w.HFONT
gdi32.GetStockObject.argtypes = (c_int,)
gdi32.GetStockObject.restype = w.HGDIOBJ

ole32 = WinDLL("ole32", use_last_error=True)
ole32.CoInitialize.argtypes = (w.LPVOID,)
ole32.CoInitialize.restype = HRESULT
ole32.CoTaskMemFree.argtypes = (w.LPVOID,)
ole32.CoTaskMemFree.restype = None

comctl32 = WinDLL("comctl32", use_last_error=True)
comctl32.InitCommonControlsEx.argtypes = (POINTER(INITCOMMONCONTROLSEX),)
comctl32.InitCommonControlsEx.restype = w.BOOL

shell32 = WinDLL("shell32", use_last_error=True)
shell32.SHBrowseForFolderA.argtypes = (POINTER(BROWSEINFOA),)
shell32.SHBrowseForFolderA.restype = POINTER(ITEMIDLIST)
shell32.SHBrowseForFolderA._winerror = _winerror
shell32.SHGetPathFromIDList.argtypes = (POINTER(ITEMIDLIST), w.LPCSTR)
shell32.SHGetPathFromIDList.restype = w.BOOL
shell32.SHGetPathFromIDList._winerror = _winerror


class Win32(GUI):
    result = ""
    pass_shown = False
    gui_display_text = ""

    start_button = None
    choose_folder_button = None
    choose_files_button = None

    input_field = None
    checkbox = None
    reveal_text = None
    label = None
    upload_label = None
    info = None
    file_info = None
    file_label = None
    progress_bar = None
    image = None

    hwnd = None
    quitting = False

    @property
    def progress(self) -> int:
        return self._progress

    @progress.setter
    def progress(self, progress: int) -> None:
        if self._closed:
            return

        if self._progress > progress:
            return

        self._progress = progress
        SendMessage(self._instance.progress_bar, PBM_SETPOS, int(self._progress), 0)

    def quit(self) -> None:
        if self._closed:
            return
        self.quitting = True
        user32.DestroyWindow(self.hwnd)

    def finish(self) -> None:
        if self._closed:
            return
        self.progress = 100
        time.sleep(1)  # give user some time to observe 100%
        self.message("Operation complete, application will close now.")

    def message(self, message: str) -> None:
        if self._closed:
            return
        user32.MessageBoxA(self.hwnd, message.encode("ascii"), b"Acquire", 0x00040000)

    def choose_folder(self) -> None:
        if self._closed:
            return

        if self.folder:
            self.folder = None
            user32.SetWindowTextA(self.label, b"No path selected...")
            user32.SetWindowTextA(self.choose_folder_button, b"Choose folder")
            user32.EnableWindow(self.choose_files_button, True)
            if not self.files:
                user32.EnableWindow(self.start_button, False)
            return

        browseinfo = BROWSEINFOA()
        browseinfo.hwndOwner = self.hwnd
        browseinfo.pidlRoot = None
        browseinfo.pszDisplayName = cast(create_string_buffer(b"", size=1000), w.LPSTR)
        browseinfo.lpszTitle = cast(create_string_buffer(b"Acquire"), w.LPCSTR)
        browseinfo.ulFlags = 0
        browseinfo.lpfn = BFFCALLBACK(_bffcallback)
        browseinfo.lParam = 0
        browseinfo.iImage = 0
        choice = shell32.SHBrowseForFolderA(byref(browseinfo))
        path = create_string_buffer(b"", size=1000)
        shell32.SHGetPathFromIDList(choice, path)
        pathstr = string_at(path).decode("utf-8")
        if pathstr:
            self.folder = Path(pathstr)
            user32.SetWindowTextA(self.label, string_at(path))
            user32.SetWindowTextA(self.choose_folder_button, b"Clear folder")
            user32.EnableWindow(self.choose_files_button, False)
            user32.EnableWindow(self.start_button, True)

        # Caller is responsible for freeing this memory.
        ole32.CoTaskMemFree(choice)

    def choose_files(self) -> None:
        if self._closed:
            return

        if self.files:
            self.files = None
            user32.SetWindowTextA(self.file_label, b"No file(s) selected...")
            user32.SetWindowTextA(self.choose_files_button, b"Choose files")
            user32.EnableWindow(self.choose_folder_button, True)
            if not self.folder:
                user32.EnableWindow(self.start_button, False)
            return

        buffer_size = 4096
        buffer = create_unicode_buffer(buffer_size)

        ofn = OPENFILENAMEW()
        ofn.lStructSize = sizeof(OPENFILENAMEW)
        ofn.hwndOwner = self.hwnd
        ofn.lpstrFile = cast(buffer, w.LPWSTR)
        ofn.nMaxFile = buffer_size
        ofn.lpstrFilter = "All Files\0*.*\0Text Files\0*.txt\0\0"
        ofn.nFilterIndex = 1
        ofn.Flags = OFN_EXPLORER | OFN_ALLOWMULTISELECT | OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST

        if comdlg32.GetOpenFileNameW(byref(ofn)):
            raw_data = string_at(buffer, buffer_size * 2).decode("utf-16le").strip("\0")
            parts = raw_data.split("\0")

            if len(parts) > 1:
                dir_path, *file_list = parts
                selected_paths = [f"{dir_path}\\{file}" for file in file_list]
            else:
                selected_paths = [parts[0]]

            if selected_paths:
                self.files = [path for path in selected_paths if path]
                user32.SetWindowTextA(self.file_label, f"{len(self.files)} file(s) selected".encode())
                user32.SetWindowTextA(self.choose_files_button, b"Clear file(s)")
                user32.EnableWindow(self.choose_folder_button, False)
                user32.EnableWindow(self.start_button, True)
        else:
            error_code = comdlg32.CommDlgExtendedError()
            if error_code == FNERR_BUFFERTOOSMALL:
                user32.MessageBoxA(
                    self.hwnd,
                    b"The combined filename length exceeded the expected size. Please select fewer files.",
                    b"Acquire",
                    0,
                )
            elif error_code != 0:
                user32.MessageBoxA(
                    self.hwnd,
                    f"Error selecting files (code: {error_code:#06x})".encode(),
                    b"Acquire",
                    0,
                )

    def show(self) -> None:
        if self._closed:
            return

        wndclass = WNDCLASSW()
        wndclass.style = CS_HREDRAW | CS_VREDRAW
        wndclass.lpfnWndProc = WNDPROC(_winmessage)
        wndclass.cbClsExtra = wndclass.cbWndExtra = 0
        wndclass.hInstance = kernel32.GetModuleHandleW(None)
        wndclass.hIcon = user32.LoadIconW(None, IDI_APPLICATION)
        wndclass.hCursor = user32.LoadCursorW(None, IDC_ARROW)
        wndclass.hbrBackground = gdi32.GetStockObject(WHITE_BRUSH)
        wndclass.lpszMenuName = None
        wndclass.lpszClassName = "AcquireGUI"
        user32.RegisterClassW(byref(wndclass))
        hwnd = user32.CreateWindowExW(
            0,
            wndclass.lpszClassName,
            "Acquire",
            WS_OVERLAPPEDWINDOW,
            CW_USEDEFAULT,
            CW_USEDEFAULT,
            600,
            400,
            None,
            None,
            wndclass.hInstance,
            None,
        )

        user32.ShowWindow(hwnd, SW_SHOWNORMAL)
        user32.UpdateWindow(hwnd)
        ole32.CoInitialize(None)
        control_list = INITCOMMONCONTROLSEX()
        control_list.dwSize = sizeof(INITCOMMONCONTROLSEX)
        control_list.dwICC = 0x00000020
        controls_loaded = comctl32.InitCommonControlsEx(byref(control_list))

        if not controls_loaded:
            raise GUIError("Unable to load GUI controls")

        hFont = gdi32.CreateFontA(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, b"Segoe UI")

        if self.upload_available:
            self.checkbox = user32.CreateWindowExW(
                0, "Button", None, WS_CHILD | WS_VISIBLE | BS_AUTOCHECKBOX, 20, 250, 16, 16, hwnd, 0, 0, 0
            )
            self.upload_label = user32.CreateWindowExW(
                0, "static", "Upload", WS_CHILD | WS_VISIBLE | SS_LEFT, 50, 250, 100, 32, hwnd, 0, 0, 0
            )
            if hFont:
                SendMessage(self.upload_label, WM_SETFONT, hFont, 1)
            GUI.auto_upload = True
            SendMessage(self.checkbox, BM_SETCHECK, 1, 0)

        self.progress_bar = user32.CreateWindowExW(
            0,
            "msctls_progress32",
            None,
            WS_CHILD | WS_VISIBLE | WS_BORDER | PBS_SMOOTH,
            20,
            300,
            550,
            32,
            hwnd,
            0,
            0,
            0,
        )

        self.info = user32.CreateWindowExW(
            0, "static", "Acquire output folder:", WS_CHILD | WS_VISIBLE, 20, 20, 200, 20, hwnd, 0, 0, 0
        )
        self.label = user32.CreateWindowExW(
            0, "static", "No path selected...", WS_CHILD | WS_VISIBLE, 20, 40, 300, 25, hwnd, 0, 0, 0
        )

        self.file_info = user32.CreateWindowExW(
            0, "static", "Select files(s) to collect:", WS_CHILD | WS_VISIBLE, 20, 130, 250, 20, hwnd, 0, 0, 0
        )

        self.choose_folder_button = user32.CreateWindowExW(
            0, "Button", "Choose folder", WS_CHILD | WS_VISIBLE | WS_BORDER | BS_FLAT, 450, 35, 120, 32, hwnd, 0, 0, 0
        )
        self.choose_files_button = user32.CreateWindowExW(
            0, "Button", "Choose files", WS_CHILD | WS_VISIBLE | WS_BORDER | BS_FLAT, 450, 145, 120, 32, hwnd, 0, 0, 0
        )

        self.file_label = user32.CreateWindowExW(
            0, "static", "No file(s) selected...", WS_CHILD | WS_VISIBLE, 20, 150, 300, 25, hwnd, 0, 0, 0
        )

        self.start_button = user32.CreateWindowExW(
            0,
            "Button",
            "Start",
            WS_CHILD | WS_VISIBLE | WS_BORDER | WS_DISABLED | BS_FLAT,
            250,
            250,
            100,
            32,
            hwnd,
            0,
            0,
            0,
        )

        if hFont:
            SendMessage(self.info, WM_SETFONT, hFont, 1)
            SendMessage(self.file_info, WM_SETFONT, hFont, 1)
            SendMessage(self.start_button, WM_SETFONT, hFont, 1)
            SendMessage(self.choose_folder_button, WM_SETFONT, hFont, 1)
            SendMessage(self.choose_files_button, WM_SETFONT, hFont, 1)
            SendMessage(self.label, WM_SETFONT, hFont, 1)
            SendMessage(self.file_label, WM_SETFONT, hFont, 1)

        msg = w.MSG()
        while user32.GetMessageW(byref(msg), None, 0, 0) != 0:
            user32.TranslateMessage(byref(msg))
            user32.DispatchMessageW(byref(msg))
            if self.quitting:
                break

    def _message(self, hwnd: w.HWND, message: w.UINT, wParam: w.WPARAM, lParam: w.LPARAM) -> w.LRESULT:
        if message == WM_COMMAND:
            if lParam == self.choose_folder_button:
                event = HIWORD(wParam)
                if event == BN_CLICKED:
                    self.choose_folder()
            elif lParam == self.choose_files_button:
                event = HIWORD(wParam)
                if event == BN_CLICKED:
                    self.choose_files()
            elif lParam == self.start_button:
                user32.EnableWindow(self.start_button, False)
                user32.EnableWindow(self.choose_folder_button, False)
                user32.EnableWindow(self.choose_files_button, False)
                if self.checkbox:
                    user32.EnableWindow(self.checkbox, False)
                self.ready = True
                self.progress = 1  # make it visible to the user that we are starting
            elif lParam == self.checkbox:
                self.auto_upload = not self.auto_upload
            return 0
        if message == WM_CLOSE:
            if self.ready:
                user32.MessageBoxA(hwnd, b"We are in the middle of acquiring this host, please wait.", b"Acquire", 0)
                return 0
            answer = user32.MessageBoxA(hwnd, b"Are you sure you want to quit?", b"Acquire", 0x01 | 0x030)
            if answer == 1:
                self._closed = True
                user32.DestroyWindow(hwnd)
            return 0

        if message == WM_CTLCOLORSTATIC and lParam in [self.upload_label, self.info, self.file_info]:
            return gdi32.GetStockObject(WHITE_BRUSH)

        if message == WM_DESTROY:
            user32.PostQuitMessage(0)
            return 0
        return user32.DefWindowProcW(hwnd, message, wParam, lParam)


# Just keep this for the function signature
def _bffcallback(hwnd: w.HWND, message: w.UINT, lParam: w.LPARAM, lpData: w.LPARAM) -> c_int:
    return 0


# Just keep this for the function signature
def _winmessage(hwnd: w.HWND, message: w.UINT, wParam: w.WPARAM, lParam: w.LPARAM) -> w.LRESULT:
    return GUI._instance._message(hwnd, message, wParam, lParam)



from __future__ import annotations

import argparse
import ctypes
import datetime
import getpass
import json
import os
import re
import sys
import textwrap
import traceback
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any

from dissect.target.helpers import keychain
from dissect.target.tools.utils.cli import _OverrideRequiredAction, list_children

from acquire.outputs import (
    COMPRESSION_METHODS,
    OUTPUTS,
    TAR_COMPRESSION_METHODS,
    ZIP_COMPRESSION_METHODS,
)

if TYPE_CHECKING:
    from dissect.target import Target

    from acquire.uploaders.plugin_registry import UploaderRegistry


class StrEnum(str, Enum):
    """Sortable and serializible string-based enum."""


def _create_profile_information(profiles: dict) -> str:
    desc = ""

    profile_names = (name for name in profiles if name != "none")
    for name in profile_names:
        profile_dict = profiles[name]
        desc += f"{name} profile:\n"

        minindent = max([len(os_) for os_ in profile_dict])
        descfmt = f"  {{:{minindent}s}}: {{}}\n"

        for os_, modlist in profile_dict.items():
            if not modlist:
                continue
            indent = 4 + len(os_)
            modlist = textwrap.wrap(", ".join([mod.__modname__ for mod in modlist]), 50)
            moddesc = modlist.pop(0)
            for ml in modlist:
                moddesc += "\n" + (" " * indent) + ml
            desc += descfmt.format(os_, moddesc)
        desc += "\n"

    return desc


def create_argument_parser(profiles: dict, volatile: dict, modules: dict) -> argparse.ArgumentParser:
    module_profiles = "Module:\n" + textwrap.indent(_create_profile_information(profiles), "  ")
    volatile_profiles = "Volatile:\n" + textwrap.indent(_create_profile_information(volatile), "  ")

    desc = module_profiles + volatile_profiles

    parser = argparse.ArgumentParser(
        prog="acquire",
        description=desc,
        epilog=(
            "If no target is specified, 'local' is used.\n\n"
            "If no options are given, the collection profile 'default' is used."
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        fromfile_prefix_chars="@",
    )

    parser.add_argument("targets", metavar="TARGETS", default=["local"], nargs="*", help="Targets to load")
    # Create a mutually exclusive group, such that only one of the output options can be used
    output_group = parser.add_mutually_exclusive_group()
    output_group.add_argument("-o", "--output", default=Path(), type=Path, help="output directory")
    output_group.add_argument("-of", "--output-file", type=Path, help="output filename")

    parser.add_argument(
        "-ot",
        "--output-type",
        choices=OUTPUTS,
        default="tar",
        help="output type (default: tar)",
    )
    parser.add_argument(
        "--compress",
        action=argparse.BooleanOptionalAction,
        help="compress output (if supported by the output type)",
    )
    parser.add_argument(
        "--compress-method",
        choices=COMPRESSION_METHODS,
        help="compression method (if supported by the output type)",
    )
    parser.add_argument(
        "--encrypt",
        action=argparse.BooleanOptionalAction,
        help="encrypt output (if supported by the output type)",
    )
    parser.add_argument(
        "--gui",
        nargs="?",
        const="always",
        default="depends",
        action="store",
        help="launch with a GUI (if available for your platform)",
    )
    parser.add_argument("--public-key", type=Path, help=argparse.SUPPRESS)
    parser.add_argument("-l", "--log", type=Path, help="log directory location")
    parser.add_argument("--no-log", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument(
        "-L",
        "--loader",
        action="store",
        default=None,
        help="select a specific loader (i.e. vmx, raw)",
    )
    parser.add_argument("-p", "--profile", choices=profiles.keys(), help="collection profile")
    parser.add_argument("--volatile-profile", choices=volatile.keys(), help="volatile profile")

    parser.add_argument(
        "--path",
        dest="path",
        action="append",
        help="acquire a path, this can be either a file or directory",
    )
    parser.add_argument("-g", "--glob", action="append", help="acquire files matching glob pattern")

    parser.add_argument("--disable-report", action="store_true", help="disable acquisition report file")

    parser.add_argument(
        "--list-children", action=_OverrideRequiredAction, help="list all children indices and paths, then exit"
    )
    parser.add_argument("--recursive", action="store_true", help="make --list-children behave recursively")

    parser.add_argument("--child", help="only collect specific child based on index or path, see --list-children")
    parser.add_argument(
        "--children",
        action=argparse.BooleanOptionalAction,
        help="collect all children in addition to main target",
    )
    parser.add_argument(
        "--skip-parent", action=argparse.BooleanOptionalAction, help="skip parent collection (when using --children)"
    )

    parser.add_argument(
        "--force-fallback",
        action=argparse.BooleanOptionalAction,
        help="force filesystem access directly through OS level. Only supported with target 'local'",
    )
    parser.add_argument(
        "--fallback",
        action=argparse.BooleanOptionalAction,
        help=(
            "fallback to OS level filesystem access if filesystem type is not supported. "
            "Only supported with target 'local'"
        ),
    )
    parser.add_argument(
        "--enable-nfs",
        action=argparse.BooleanOptionalAction,
        help="mount nfs shares by connecting to the nfs server of the share.Only supported with target 'local'",
    )

    parser.add_argument(
        "-u",
        "--auto-upload",
        action=argparse.BooleanOptionalAction,
        help="upload result files after collection",
    )
    parser.add_argument(
        "--upload",
        nargs="+",
        help="upload specified files (all other acquire actions are ignored)",
    )
    parser.add_argument("--no-proxy", action="store_true", help="don't autodetect proxies")

    parser.add_argument("-K", "--keychain-file", type=Path, help="keychain file in CSV format")
    parser.add_argument("-Kv", "--keychain-value", help="passphrase, recovery key or key file path value")

    for module_cls in modules.values():
        for args, kwargs in module_cls.__cli_args__:
            parser.add_argument(*args, **kwargs)

    parser.add_argument("-v", "--verbose", action="count", default=3, help="increase output verbosity")
    return parser


def parse_acquire_args(
    parser: argparse.ArgumentParser,
    config: dict[str, Any],
) -> tuple[argparse.Namespace, list[str]]:
    """Parse and set the acquire command line arguments.

    The arguments are set to values supplied in ``config[arguments]``, when not
    changed from the default values specified in ``parser``.

    The ``config`` dict is added to the parsed command line arguments for
    convenience of later use.

    Args:
        parser: A parser for acquire command line arguments.
        config: A dict of global configuration values.

    Returns:
        A command line arguments namespace
    """
    args, rest = parser.parse_known_args()
    _merge_args_and_config(parser, args, config)

    if args.list_children:
        # List found children on targets and exit
        list_children(args)
        parser.exit(0)

    return args, rest


def _merge_args_and_config(
    parser: argparse.ArgumentParser,
    command_line_args: argparse.Namespace,
    config: dict[str, Any],
) -> None:
    """Update the parsed command line arguments with the optional set of configured default arguments.

    The arguments are set to values supplied in ``config[arguments]``, when not
    changed from the default values specified in ``parser``.

    The ``config`` dict is added to the parsed command line arguments for
    convenience of later use.

    Args:
        parser: A parser for acquire command line arguments.
        command_line_args: The namespace with parsed commandline args. This
                           namespace is updated with the optional set of
                           configured default arguments.
        config: A dict of global configuration values.
    """
    config_defaults = config.get("arguments")
    if not config_defaults:
        config_defaults = config["arguments"] = []

    config_defaults_args = parser.parse_args(config_defaults)

    for argument, value in command_line_args._get_kwargs():
        if parser.get_default(argument) == value:
            config_argument = getattr(config_defaults_args, argument, value)
            setattr(command_line_args, argument, config_argument)

    command_line_args.config = config


def check_and_set_log_args(args: argparse.Namespace) -> None:
    """Check command line arguments which are related to logging.

    Also some arguments derived from the user supplied ones are set in the
    ``args`` namespace for convenience.

    This function is separate from ``check_and_set_acquire_args()`` as logging
    needs to be setup as soon as possible when running acquire.

    Args:
        args: The namespace containing the command line arguments.

    Raises:
        ValueError: When an invalid combination of arguments is found.
    """
    start_time = get_utc_now_str()
    log_path = None
    log_to_dir = False
    log_delay = False

    if not args.no_log:
        log_path = args.log or args.output or args.output_file.parent

        if log_path.is_dir():
            log_to_dir = True
            log_delay = True
        elif log_path.is_file() or (not log_path.exists() and log_path.parent.is_dir()):
            # Logging to a single file is allowed, even if the file does not yet
            # exist, as it will be automatically created. However then the parent
            # directory must exist.
            if args.children or len(args.targets) > 1:
                # If children or multiple targets are acquired, logging can only happen to separate
                # files, so log_path needs to be a directory.
                raise ValueError("Log path must be a directory when using multiple targets or --children")
        else:
            raise ValueError(f"Log path doesn't exist: {log_path}")

    args.start_time = start_time
    args.log_path = log_path
    args.log_to_dir = log_to_dir
    args.log_delay = log_delay


def check_and_set_acquire_args(
    args: argparse.Namespace,
    upload_plugins: UploaderRegistry,
) -> None:
    """Check the command line arguments and set some derived arguments.

    This function is separate from ``check_and_set_log_args()`` as logging
    needs to be setup as soon as possible when running acquire.

    Args:
        args: The namespace containing the command line arguments.
        upload_plugins: The registry of available upload plugins.

    Raises:
        ValueError: When an invalid combination of arguments is found.
    """
    upload_plugin = None
    args.upload_plugin = upload_plugin

    # check & set upload related configuration
    if args.upload and args.auto_upload:
        raise ValueError("only one of --upload or --auto-upload can be specified")

    if args.upload or args.auto_upload:
        upload_mode = args.config.get("upload", {}).get("mode")
        if not upload_mode:
            raise ValueError("Uploading is not configured")

        upload_plugin_cls = upload_plugins.get(upload_mode)
        if not upload_plugin_cls:
            raise ValueError(f"Invalid upload mode: {upload_mode}")

        # If initialization of the plugin fails, a ValueError is raised
        upload_plugin = upload_plugin_cls(**args.config)
        args.upload_plugin = upload_plugin

    if not args.upload:
        # check output related configuration
        if (args.children or len(args.targets) > 1) and args.output_file:
            raise ValueError("--children can not be used with --output-file. Use --output instead")
        if args.output_file and (not args.output_file.parent.is_dir() or args.output_file.is_dir()):
            raise ValueError("--output-file must be a path to a file in an existing directory")
        if args.output and not args.output.is_dir():
            raise ValueError(f"Output directory doesn't exist or is a file: {args.output}")

        # check & set encryption related configuration
        if args.encrypt:
            public_key = args.config.get("public_key")
            if not public_key and args.public_key and args.public_key.is_file():
                public_key = args.public_key.read_text()
            if not public_key:
                raise ValueError("No public key available (embedded or argument)")
            args.public_key = public_key

    if not args.children and args.skip_parent:
        raise ValueError("--skip-parent can only be set with --children")

    if args.compress:
        if (args.output_type == "zip" and args.compress_method) and args.compress_method not in ZIP_COMPRESSION_METHODS:
            raise ValueError(
                f"Invalid compression method for zip, allowed are: {', '.join(ZIP_COMPRESSION_METHODS.keys())}"
            )
        if (args.output_type == "tar" and args.compress_method) and args.compress_method not in TAR_COMPRESSION_METHODS:
            raise ValueError(
                f"Invalid compression method for tar, allowed are: {', '.join(TAR_COMPRESSION_METHODS.keys())}"
            )

    if args.keychain_file:
        keychain.register_keychain_file(args.keychain_file)

    if args.keychain_value:
        keychain.register_wildcard_value(args.keychain_value)


def get_user_name() -> str:
    try:
        return getpass.getuser()
    except Exception:
        return "unknown"


def is_user_admin() -> bool:
    try:
        return os.geteuid() == 0
    except AttributeError:
        pass

    try:
        return ctypes.windll.shell32.IsUserAnAdmin() == 1
    except Exception:
        return False


def get_utc_now() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)


def get_utc_now_str() -> str:
    return get_utc_now().strftime("%Y%m%d%H%M%S")


def get_formatted_exception() -> str:
    exc_info = sys.exc_info()
    return "".join(traceback.format_exception(*exc_info))


def format_output_name(prefix: str, postfix: str | None = None, ext: str | None = None) -> str:
    if not postfix:
        postfix = datetime.datetime.now(tz=datetime.timezone.utc).strftime("%Y%m%d%H%M%S")
    name = f"{prefix}_{postfix}"
    if ext:
        name = f"{name}.{ext}"
    return name


def persist_execution_report(path: Path, report_data: dict) -> Path:
    with path.open("w") as f:
        f.write(json.dumps(report_data, sort_keys=True, indent=4))


DEVICE_SUBST = re.compile(r"^(/\?\?/)")
SYSVOL_SUBST = re.compile(r"^/?sysvol(?=/|$)", flags=re.IGNORECASE)

SYSVOL_UPPER_SUBST = re.compile(r"^(/?SYSVOL)(?=/|$)")
DRIVE_LOWER_SUBST = re.compile(r"^(/?[a-z]:)(?=/|$)")


def normalize_path(
    target: Target,
    path: str | Path,
    resolve_parents: bool = False,
    preserve_case: bool = True,
) -> str:
    if isinstance(path, Path):
        if resolve_parents:
            path = path.parent.resolve().joinpath(path.name)

        path = path.as_posix()

    if target.os == "windows":
        path = DEVICE_SUBST.sub("", path)
        if sysvol_drive := target.props.get("sysvol_drive"):
            path = SYSVOL_SUBST.sub(sysvol_drive, path)

        # The substitutions below are temporary until we have proper full path name uniformization
        # for case insensitive filesystems.
        # Replace any uppercase SYSVOL path, with a lowercase version.
        path = SYSVOL_UPPER_SUBST.sub(lambda pat: pat.group(1).lower(), path)
        # Replace any lower case driveletter path with an uppercase version.
        path = DRIVE_LOWER_SUBST.sub(lambda pat: pat.group(1).upper(), path)

    if not target.fs.case_sensitive and not preserve_case:
        path = path.lower()

    return path
